# -*- coding:utf-8 -*-
from elixir import Entity, ManyToOne, OneToMany, UnicodeText, Field, Binary, Boolean, Text, Integer
from os.path import join
from os import unlink
from arv.lib.util import (valid, try_unique_column, get_keyid_in_certif,
    get_keyid_from_certifstring, trace, bin_encoding, suppress_colon)
from creole import cert
from arv.config import (x509_default_key_bits, ssl_default_cert_time,
    ca_conf_file, conf_file_orig, ssl_dir)
from arv.lib.logger import logger

class TmplEdge(Entity):
    # filled at init time
    name = Field(UnicodeText, unique=True)
    tmpl_vertex_a = ManyToOne('TmplVertex')
    tmpl_vertex_b = ManyToOne('TmplVertex')
    tmpl_connect = ManyToOne('TmplConnect')
    # filled later
    edges = OneToMany('Edge')
    state = Field(Integer, default=0)

    def add_edge(self, connect):
        try:
            return Edge(tmpl_edge=self, connect=connect)
        except Exception, e:
            raise ValueError(str(e))

    def mod_name(self, name):
        name = valid(name, 'string')
        if name != self.name:
            if TmplEdge.query.filter_by(name=name).first() != None:
                raise Exception('Name {0} already used'.format(name))
            self.name = name

    def mod_tmpl_vertex1(self, tmpl_vertex1):
        if self.tmpl_vertex_a !=  tmpl_vertex1:
            if self.edges != []:
                raise Exception('This TmplEdge is already used, cannot modify tmpl_vertex')
            self.tmpl_vertex_a = tmpl_vertex1

    def mod_tmpl_vertex2(self, tmpl_vertex2):
        if self.tmpl_vertex_b !=  tmpl_vertex2:
            if self.edges != []:
                raise Exception('This TmplEdge is already used, cannot modify tmpl_vertex')
            self.tmpl_vertex_b = tmpl_vertex2

@trace
def get_tmpl_edges(tmpl_connect):
    try:
        return TmplEdge.query.filter_by(tmpl_connect=tmpl_connect).all()
    except:
        raise Exception( 'error in get_tmpl_edges')

@trace
def del_tmpl_edge(tmpl_edge):
    if Edge.query.filter_by(tmpl_edge=tmpl_edge).first():
        raise Exception("TmplEdge is already used by a Edge, you cannot delete it")
    try:
        #removes vertex_a if only used by itself
        if tmpl_edge.tmpl_vertex_a.tmpl_edges_a == [tmpl_edge]:
            tmpl_edge.tmpl_vertex_a.delete()
        if tmpl_edge.tmpl_vertex_b.tmpl_edges_b == [tmpl_edge]:
            tmpl_edge.tmpl_vertex_b.delete()
    except:
        raise Exception("error when deleting tmpl_vertices")
    try:
        tmpl_edge.delete()
    except:
        raise Exception("error in del_tmpl_edge")

class TmplConnect(Entity):
    # filled at init time
    name = Field(UnicodeText, unique=True)
    tmpl_node_a = ManyToOne('TmplNode', required=True)
    tmpl_node_b = ManyToOne('TmplNode', required=True)
    # filled later
    tmpl_edges = OneToMany('TmplEdge')
    connects = OneToMany('Connect')
    cred_auth = ManyToOne('Credential')
    state = Field(Integer, default=0)

    def _add_tmpl_edge(self, name, tmpl_vertex_a, tmpl_vertex_b):
        name = valid(name, 'string')
        tmpl = TmplEdge(name=name, tmpl_vertex_a=tmpl_vertex_a, tmpl_vertex_b=tmpl_vertex_b, tmpl_connect=self)
        self.tmpl_edges.append(tmpl)
        return tmpl

    def add_tmpl_edge(self, name, tmpl_vertex_a, tmpl_vertex_b):
        if TmplEdge.query.filter_by(tmpl_vertex_a=tmpl_vertex_a, tmpl_vertex_b=tmpl_vertex_b, tmpl_connect=self).first() != None:
            raise Exception('TmplEdge already exists')
        return try_unique_column('add_tmpl_edge', self._add_tmpl_edge,
                name=name, tmpl_vertex_a=tmpl_vertex_a,
                tmpl_vertex_b=tmpl_vertex_b)
#        try:
#            name = valid(name, 'string')
#            tmpl = TmplEdge(name=name, tmpl_vertex_a=tmpl_vertex_a, tmpl_vertex_b=tmpl_vertex_b, tmpl_connect=self)
#            self.tmpl_edges.append(tmpl)
#        except IntegrityError, e:
#            if " column name is not unique " in str(e):
#                raise Exception("Name should be unique")
#            else:
#                raise Exception("error in add_tmpl_edge")
#        except:
#            raise Exception("error in add_tmpl_edge")
#        return tmpl

    def add_connect(self, tail_node, head_node, tail_extr, head_extr, tail_cred, head_cred):
        try:
            if tail_node == head_node or tail_extr == head_extr:
                raise Exception( "wrong argument on add_connect")
            return Connect(tmpl_connect=self, tail_node=tail_node,
                    head_node=head_node, tail_extr=tail_extr,
                    head_extr=head_extr, tail_cred=tail_cred,
                    head_cred=head_cred)
        except:
            raise Exception("error in add_connect")

@trace
def get_local_ca():
    return Credential.query.filter_by(ca=True, local=True).first()

@trace
def get_credentials_auth():
    try:
        return Credential.query.filter_by(ca=True).all()
    except:
        raise Exception("error in get_credentials_auth")

@trace
def get_credential_auth(credential):
    try:
        return Credential.query.filter_by(credential=credential).first()
    except:
        return None

@trace
def add_credential_auth(credential, passwd=None, local=False):
    try:
        subject = cert.get_subject(certfile=credential)
        name = unicode(subject[1], 'utf-8')
        suffix_cred = subject[0]
        suffix_cred = valid(suffix_cred, 'string')
        subject = bin_encoding("%s/CN=%s"%(suffix_cred, name))
        name = valid(name, 'string')
        if local:
            # get the keyid from the private key
            fh = open(cert.ca_keyfile, 'r')
            private_key = fh.read()
            fh.close()
            subjkey, keyid_cred = get_keyid_from_certifstring(private_key, passwd, mode='rsa')
            subjkey = suppress_colon(subjkey)
            keyid_cred = suppress_colon(keyid_cred)
            return Credential(name=name, suffix_cred=suffix_cred, subject=subject,
                              private_key=private_key, credential=credential,
                              ca=True, local=local, keyid_cred=keyid_cred, subjkey=subjkey)
        else:
            #FIXME: calculer les subjkey et keyid
            subjkey, keyid_cred = get_keyid_from_certifstring(credential, mode='x509')
            subjkey = suppress_colon(subjkey)
            keyid_cred = suppress_colon(keyid_cred)
            return Credential(name=name, suffix_cred=suffix_cred, subject=subject,
                              credential=credential, ca=True, local=local,
                              keyid_cred=keyid_cred, subjkey=subjkey)

    except Exception, e:
        raise Exception("error in add_cred_auth: %s" % e)

@trace
def add_tmpl_connect(name, node_a, node_b, cred_auth):
    try:
        name = valid(name, 'string')
        if not cred_auth.ca:
            raise Exception('add_tmpl_connect: attribute cred_auth not a credential authority')
    except Exception, e:
        raise ValueError(str(e))
    if TmplConnect.query.filter_by(name=name).first() != None:
        raise Exception("Name should be unique")
    try:
        return TmplConnect(name=name, tmpl_node_a=node_a, tmpl_node_b=node_b, cred_auth=cred_auth)
    except:
        raise Exception("error in add_tmpl_connect")

@trace
def get_tmpl_connects(tmpl_node_a=None, tmpl_node_b=None):
    try:
        if tmpl_node_a == None:
            return TmplConnect.query.all()
        else:
            #tmpl_node_a could be node_a or node_b and tmpl_node_b could be node_a or node_b
            tmpl_connects = set(TmplConnect.query.filter_by(tmpl_node_a=tmpl_node_a,
                    tmpl_node_b=tmpl_node_b).all())
            union = tmpl_connects | set(TmplConnect.query.filter_by(tmpl_node_b=tmpl_node_a,
                    tmpl_node_a=tmpl_node_b).all())
            return list(union)
    except Exception, e:
        raise Exception( 'error in get_tmpl_connects : %s '% str(e))

@trace
def del_tmpl_connect(tmpl_connect):
    if Connect.query.filter_by(tmpl_connect=tmpl_connect).first():
        raise Exception("TmplConnect is already used by a Connect, you cannot delete it")
    try:
        for tmpledge in tmpl_connect.tmpl_edges:
            del_tmpl_edge(tmpledge)
    except:
        import traceback
        traceback.print_exc()
        raise Exception("error when delete tmpledges")
    try:
        tmpl_connect.delete()
    except:
        raise Exception("error in del_tmpl_connect")

# ____________________________________________________________

class Edge(Entity):
    # filled at init time
    tmpl_edge = ManyToOne('TmplEdge')
    connect = ManyToOne('Connect')
    state = Field(Integer, default=0)

@trace
def get_edges(nodea=None, nodeb=None, tmpl_connect=None):
    if nodea == None and nodeb == None and tmpl_connect == None:
        return Edge.query.all()
    tmpl_node_a = nodea.tmpl_node
    tmpl_node_b = nodeb.tmpl_node
    edges=[]
    if tmpl_connect.edges != None:
        for tmpledge in tmpl_connect.edges:
            #nodea nodeb
            edges.extend(tmpledge.edges)
    return edges

@trace
def del_edge(edge):
    try:
        edge.delete()
    except:
        raise Exception("error in del_edge")

class Connect(Entity):
    # filled at init time
    tail_node = ManyToOne('Node')
    head_node = ManyToOne('Node')
    tail_extr = ManyToOne('Extremity')
    head_extr = ManyToOne('Extremity')
    tail_cred = ManyToOne('Credential')
    head_cred = ManyToOne('Credential')
    tmpl_connect = ManyToOne('TmplConnect', required=True)
    edges = OneToMany('Edge')
    state = Field(Integer, default=0)

    def mod_tail_extr(self, extr):
        if extr != self.tail_extr:
            self.tail_extr = extr

    def mod_head_extr(self, extr):
        if extr != self.head_extr:
            self.head_extr = extr

    def mod_tail_cred(self, cred):
        if cred != self.tail_cred:
            self.tail_cred = cred

    def mod_head_cred(self, cred):
        if cred != self.head_cred:
            self.head_cred = cred

@trace
def get_connects(nodea=None, nodeb=None):
    try:
        if nodea == None and nodeb == None:
            return Connect.query.all()
        if nodea == None:
            connect = Connect.query.filter_by(head_node=nodeb).all()
            connect.extend(Connect.query.filter_by(tail_node=nodeb).all())
            return connect
        if nodeb == None:
            connect = Connect.query.filter_by(tail_node=nodea).all()
            connect.extend(Connect.query.filter_by(head_node=nodea).all())
            return connect
        connect = Connect.query.filter_by(tail_node=nodea, head_node=nodeb).all()
        connect.extend(Connect.query.filter_by(head_node=nodea, tail_node=nodeb).all())
        return connect
    except Exception, e:
        raise ValueError('error in get_connects: '+str(e))

@trace
def get_connected_nodes(nodea):
    try:
        nodeb = []
        connects = Connect.query.filter_by(head_node=nodea).all()
        for connect in connects:
            nodeb.append(connect.tail_node)
        connects = Connect.query.filter_by(tail_node=nodea).all()
        for connect in connects:
            nodeb.append(connect.head_node)
        return nodeb
    except Exception, e:
        raise ValueError('error in get_connected_nodes: '+str(e))

@trace
def del_connect(connect):
    try:
        connect.delete()
    except:
        raise Exception("error in del_connect")


class Credential(Entity):
    # filled at init time
    name = Field(Text, unique=True)
    credential = Field(Binary)
    unsigned_credential = Field(Binary)
    # only if ca = False
    private_key = Field(Binary, unique=True)
    node = ManyToOne('Node')
    ca = Field(Boolean)
    # only if ca = True
    suffix_cred = Field(Text)
    subject = Field(Binary, unique=True)
    local = Field(Boolean, default=False)
    #Ajouter un champ crl
    clr_uris = OneToMany('Crl')
    # filled later
    ## only if ca = False
    cred_auth = ManyToOne('Credential', inverse='credentials')
    ## only if ca = True
    keyid_cred = Field(Binary)
    subjkey = Field(Binary)
    tmpl_connects = OneToMany('TmplConnect')
    credentials = OneToMany('Credential')
    head_conns = OneToMany('Connect', inverse='head_cred')
    tail_conns = OneToMany('Connect', inverse='tail_cred')
    state = Field(Integer, default=0)

#Table Crl: liste des crl d'une CA
class Crl(Entity):
    uri = Field(Text)
    ca = ManyToOne('Credential')

@trace
def add_crl_uri(cred_auth, uri):
    return Crl(ca=cred_auth, uri=uri)

@trace
def get_credential_by_ca_node(cred_auth, node):
    return Credential.query.filter_by(cred_auth=cred_auth, node=node).first()

@trace
def get_credential(name):
    return Credential.query.filter_by(name=name).first()

@trace
def add_credential(name, passwd, node, cred_priv_type):
    if Credential.query.filter_by(name=name).first() != None:
        raise Exception('Name should be unique')
    uname = valid(name, 'string')
    #passwd  = valid(passwd, 'string')
    #passwd = str(passwd)
    conf_file = '/tmp/%s.conf'%name
    certfile = join(ssl_dir, 'certs', '%s.pem'%name)
    keyfile = join(ssl_dir, 'private', 'priv-%s.pem'%name)
    reqfile = join(ssl_dir, 'req', '%s.p10'%name)
    #copy(conf_file_orig, conf_file)
    fh = open(conf_file_orig, 'r')
    content = fh.read()
    fh.close()
    content = content.replace('@@CN-SERVEUR@@', name)
    content = content.replace('@@PASSWD@@', passwd)
    if node.tmpl_node.mimetype == u'sphynx':
        content = content.replace('encrypt_key\t\t= yes', 'encrypt_key\t\t= no')
    fh = open(conf_file, 'w')
    logger.debug('%s content: %s'%(conf_file, content))
    fh.write(content)
    fh.close()
    cert.load_conf({'ssl_dir': ssl_dir, 'start_index': "01", 'conf_file': conf_file,
        'ca_conf_file': ca_conf_file, 'ssl_default_cert_time': ssl_default_cert_time,
        'ssl_default_key_bits': x509_default_key_bits})
    if cred_priv_type == 'autosigned':
        cert.gen_certif(certfile=certfile,
                        keyfile=keyfile,
                        del_passwd=False)

        fh = open(certfile, 'r')
        credential = fh.read()
        fh.close()
        fh = open(keyfile, 'r')
        private_key = fh.read()
        fh.close()
        local_ca = get_local_ca()
        if local_ca == None:
            raise Exception('No local CA found')
        # get the keyid from the certificate
        subjkey, keyid_cred = get_keyid_in_certif(certfile, passwd=passwd, certiftype='notca')
        subjkey = suppress_colon(subjkey)
        keyid_cred = suppress_colon(keyid_cred)
        node.add_credential(name=uname,
                            private_key=private_key,
                            credential=credential,
                            cred_auth=local_ca,
                            keyid_cred=keyid_cred,
                            subjkey=subjkey)
        unlink(certfile)
        unlink(keyfile)
        unlink(reqfile)
        unlink(conf_file)
        return True
    elif cred_priv_type == 'cred_request':
        cert.gen_certif(certfile=certfile,
                        keyfile=keyfile,
                        del_passwd=False,
                        signe_req=False)
        fh = open(keyfile, 'r')
        private_key = fh.read()
        fh.close()
        fh = open(reqfile, 'r')
        req_credential = fh.read()
        fh.close()
        unlink(conf_file)
        return req_credential
    else:
        raise Exception('Unknown cred_priv_type %s, should be autosigned or cred_request'%cred_priv_type)

    unlink(conf_file)

