# -*- coding:utf-8 -*-
import re
from tempfile import mkstemp
from os import unlink, system
from os.path import basename, dirname, join
from commands import getoutput, getstatusoutput

from creole.cert import ca_keyfile

from formencode import validators, foreach
from sqlalchemy.exc import IntegrityError
from arv.config import key2keyid_path, id2sql_path, bin2sql_path, ssl_dir
from arv.lib.logger import logger

def trace(func):
    """This is a decorator which can be used to trace functions
    """
    def newFunc(*args, **kwargs):
        logger.debug("-> entering %s(%s, %s)"% (func.__name__, str(args), str(kwargs)))
        return func(*args, **kwargs)
    newFunc.__name__ = func.__name__
    newFunc.__doc__ = func.__doc__
    newFunc.__dict__.update(func.__dict__)
    return newFunc

@trace
def normalize_unicode(string):
    if not type(string) == unicode:
        if type(string) == str:
            string = unicode(string, "utf-8")
        else:
            raise TypeError( "unsupported encoding")
    return string

@trace
def valid(value, typ):
    """
        formencode validation
    """
    if typ == 'string':
        validator = validators.String()
    if typ == 'bool':
        validator = validators.StringBoolean()
    if typ == 'integer':
        validator = validators.Int()
    if typ == 'enum':
        #FIXME : nothing is done in this case by now
        #validator = foreach.ForEach()
        return value
    if typ == 'ip':
        validator = validators.IPAddress()
    val = validator.to_python(str(value))
    if typ == 'string' or typ == 'ip':
        val = normalize_unicode(val)
    return val
# ____________________________________________________________
@trace
def try_unique_column(func_name, function, **args):
    try:
        return function(**args)
    except IntegrityError, e:
        if " column name is not unique " in str(e):
            raise Exception("Name should be unique")
        else:
            raise Exception("error in %s: %s" % (func_name, str(e)))
    except Exception, e:
        raise Exception("error in %s: %s" % (func_name, str(e)))

# ____________________________________________________________
#
@trace
def get_scndline(lines):
    """parse a two lines (type, name) from a command output
    typically :

      type    encoding
      2,  X'736466736466'
    """
    logger.debug("scndline : " + lines)
    second_line = lines.split('\n')[1]
    keyid = second_line.split(',')[1]
    encoded_string = keyid.strip()
    #encoded_string =  encoded_string.replace("X'", "")
    #encoded_string = encoded_string.replace("'", "")
    if encoded_string.startswith("X'") and encoded_string.endswith("'"):
        return encoded_string[2:-1]
    else:
        raise TypeError("unexpected encoded string: {0}".format(encoded_string))

@trace
def get_lastline(lines):
    """parse lines (type, name) from a command output
    typically :

    writing RSA key
    parsed 2048 bits RSA private key.
    subjectPublicKeyInfo keyid: b0:71:fb:0a:62:f7:8d:7b:9d:35:d7:c9:4d:12:f5:d5:51:e6:db:da
    subjectPublicKey keyid:     0e:5d:33:43:15:b0:f2:a0:0b:b6:6b:f3:24:44:6a:f5:08:91:da:0d

    """
    logger.debug("lastline : " + lines)
    lines = lines.split('\n')
    last_line = lines[-1]
    lines.pop()
    penultimate = lines[-1]
    subjkey = penultimate.replace('subjectPublicKeyInfo keyid:', '')
    # impossible : so much semicolumnsi in the keyid itself
    # keyid = last_line.split(':')[1]
    keyid = last_line.replace('subjectPublicKey keyid:', '')
    return (subjkey.strip(), keyid.strip())

@trace
def bin_encoding(clear):
    """encode in binary for strongswan database
    """
    cmd = '%s "%s"' % (id2sql_path, clear)
    result = get_scndline(getoutput(cmd))
    #encode and decode. this functionality already exists with the encodings library (which is built-in)
    # have to decode to put the normal string in a sqlalchemy binary type
    logger.debug("#-> bin_encoding result var: " + result)
    return result.decode("hex")

@trace
def suppress_colon(keyid):
    """suppress colons in a string
    """
    return keyid.replace(':', '').decode('hex')

@trace
def get_keyid_in_certif(certif_name, passwd=None, certiftype='ca'):
    """takes the long name (ex: /var/lib/arv/CA/certs/CaCert.pem) of the certificate
    and returns the keyid
    """
    certif_path = dirname(certif_name)
    certif_name = basename(certif_name)
    certif_path = dirname(certif_path)
    certif_path = join(certif_path, 'private')
    if certiftype == 'ca':
        priv_certname = ca_keyfile
    else:
        priv_certname = join(certif_path, 'priv-'+certif_name)
    return get_keyid_from_keyid_in_certif(priv_certname, passwd, mode='rsa')

@trace
def get_keyid_from_keyid_in_certif(certif_name, passwd=None, mode='rsa'):
    """constructs a keyid when a keyid is not present in the certificate
    """
    if mode == 'rsa':
        #Extract from private key
        cmd = 'openssl {0} -in {1} -text -passin pass:"{2}" | {3}'.format(mode, certif_name, passwd, key2keyid_path)
        subjkey, keyid = get_lastline(getoutput(cmd))
    elif mode == 'x509':
        #Extract from credential
        cmd = 'openssl {0} -in {1} -text -noout| grep keyid'.format(mode, certif_name)
        output = getoutput(cmd)
        keyid = output.split("keyid:")[1]
        logger.debug('{0}'.format(keyid))
        subjkey = keyid
    return (subjkey, keyid)

@trace
def get_keyid_from_certifstring(certif_string, passwd=None, mode='rsa'):
    """constructs a keyid when a keyid is not present in the certificate
    """
    try:
        fd, certif_name = mkstemp()
        fh = file(certif_name, 'w')
        fh.write(certif_string)
        fh.close()
        subjkey, keyid = get_keyid_from_keyid_in_certif(certif_name, passwd, mode=mode)
        unlink(certif_name)
        return (subjkey, keyid)
    except Exception, e:
        logger.warning('cannot generate keyid in certificate %s: %s'% (certif_string, str(e)))
        raise Exception('Cannot generate keyid in this certificate')

@trace
def test_password(private_key, password):
    retcode = system("""echo "{0}" | openssl rsa -text -passin pass:"{1}" -noout &> /dev/null""".format(private_key, password))
    if retcode == 0:
        return True
    else:
        return False

@trace
def decrypt_privkey(privkey_string, passwd):
    """Suppress password from private key
    """
    cmd = 'echo -e "{0}" | openssl rsa -passin pass:"{1}" 2>&1'.format(privkey_string, passwd)
    errcode, output = getstatusoutput(cmd)
    if errcode:
        raise Exception('Unable to decrypt private key, check password')
    lines = output.split('\n')
    lines = lines[1:]
    decrypted_key = "\n".join(lines)
    return decrypted_key

def get_req_archive(name):
    """Download credential request to send to CA
    """
    reqname = name +".p10"
    privkeyname = "priv-"+name+".pem"
    reqfilename = join(ssl_dir, "req",  reqname)
    privkeyfilename = join(ssl_dir, "private", privkeyname)
    system("""cd {0}
mv {1} {0}
mv {2} {0}
tar -czf {3}.tgz {4} {5}
rm {4} {5}
""".format(ssl_dir, reqfilename, privkeyfilename, name, reqname, privkeyname))
    tarfname = join(ssl_dir, name+".tgz")
    fh = file(tarfname, 'r')
    content = fh.read()
    fh.close()
    unlink(tarfname)
    return content

@trace
def split_pkcs7(pkcs7_cred):
    """Split pkcs7 string in two strings ca_cred and cred
    """
    try:
        cmd = 'echo "{0}" | openssl pkcs7 -print_certs '.format(pkcs7_cred)
        two_creds = getoutput(cmd)
        # = content.slpit('\n')
        new_cred = True
        issuer = ''
        ca_cred = ''
        cred = ''
        for line in two_creds.splitlines():
            if line == '':
                subject = ''
                issuer = ''
                new_cred = True
            elif new_cred:
                subject = re.match( r'subject(.*)CN=', line)
                if not subject:
                    raise Exception('Cannot get subject from pkcs7 file')
                subject = subject.group(1)
                new_cred = False
            elif not new_cred and issuer == '' and subject:
                issuer = re.match( r'issuer(.*)CN=', line)
                if not issuer:
                    raise Exception('Cannot get issuer from pkcs7 file')
                issuer = issuer.group(1)
                if subject == issuer:
                    if ca_cred != '':
                        raise Exception('Already a ca_cred in pkcs7 file')
                else:
                    if cred != '':
                        raise Exception('Already a cred in pkcs7 file')
            elif subject and issuer:
                if subject == issuer:
                    ca_cred += '{0}\n'.format(line)
                else:
                    cred += '{0}\n'.format(line)
            else:
                raise Exception('Unknown line in pkcs7 file {0}'.format(line))

        if ca_cred == '' or cred == '':
            raise Exception('Pkcs7 file not complete')
        return (ca_cred, cred)
    except Exception, e:
        logger.error('Cannot read pkcs7: {1}'.format(e))
        raise Exception('Cannot read pkcs7: {1}'.format(e))

@trace
def extract_crls_from_certifstring(certifstring):
    """Extract crl from certif string
    """
    try:
        cmd = 'echo "{0}"|openssl x509 -text -noout|grep "URI"'.format(certifstring)
        errcode, output = getstatusoutput(cmd)
        output = output.split('\n')
        crl = []
        crlbegin = False
        for line in output:
            if "CRL Distribution Points:" in line:
                crlbegin = True
                continue
            if crlbegin:
                try:
                    crl.append(line.split('URI:')[1])
                except:
                    break
        return crl
    except:
        logger.warning("No crl found in credential")
        pass


